#! /bin/env python

import struct
import socket
import time
import sys
import os
import bp
import logger
import traceback

import params

import gevent
from gevent.queue import Queue
from gevent.event import AsyncResult

MESSAGE_MAGIC               = 'CB'
MESSAGE_VER                 = 0x01
MESSAGE_TYPE_WELCOME        = 0x01
MESSAGE_TYPE_CLOSE          = 0x02
MESSAGE_TYPE_QUERY          = 0x03
MESSAGE_TYPE_ANSWER         = 0x04

class Messager:

    @staticmethod
    def data_for_welcome():
        return struct.pack('<2s2B', MESSAGE_MAGIC, MESSAGE_VER, MESSAGE_TYPE_WELCOME)

    @staticmethod
    def data_for_close():
        return struct.pack('<2s2B', MESSAGE_MAGIC, MESSAGE_VER, MESSAGE_TYPE_WELCOME)

    @staticmethod
    def data_for_query(qid, service, method, params):
        s = bp.encode((qid, service, method, params))
        header = struct.pack('<2s2Bi', MESSAGE_MAGIC, MESSAGE_VER, MESSAGE_TYPE_QUERY, len(s))
        return header + str(s)

    @staticmethod
    def data_for_answer(qid, status, data):
        s = bp.encode((qid, status, data))
        header = struct.pack('<2s2Bi', MESSAGE_MAGIC, MESSAGE_VER, MESSAGE_TYPE_ANSWER, len(s))
        return header + str(s)

    @staticmethod
    def receive_data(socket, num):
        data = ''
        while len(data) < num:
            buf = socket.recv(num - len(data))
            if not buf:
                return None
            data += buf
        return data

    @staticmethod
    def receive_msg_type(socket):
        header_str = Messager.receive_data(socket, 4)
        if not header_str:
            return None

        header = struct.unpack('<2s2B', header_str)
        if header[0] != MESSAGE_MAGIC or header[1] != MESSAGE_VER:
            return None
        return header[2]

    @staticmethod
    def receive_msg(socket):
        type = Messager.receive_msg_type(socket)

        if not type:
            return None

        if type == MESSAGE_TYPE_WELCOME:
            return Welcome()
        elif type == MESSAGE_TYPE_CLOSE:
            return Close()

        len_str = Messager.receive_data(socket, 4)
        if not len_str:
            return None
        len = struct.unpack('<i', len_str)
        len = len[0]

        body = Messager.receive_data(socket, len)
        data = bp.decode(body)

        if type == MESSAGE_TYPE_QUERY:
            return Query.from_receive(data)
        elif type == MESSAGE_TYPE_ANSWER:
            return Answer.from_receive(data)
        else:
            return None

    @staticmethod
    def message_from_data(data):
        if len(data) < 4:
            return None

        header = struct.unpack('<2s2Bi', data[:4])
        if header[0] != MESSAGE_MAGIC or header[1] != MESSAGE_VER:
            return None
        type = header[2]

        if type == MESSAGE_TYPE_WELCOME:
            return Welcome()
        elif type == MESSAGE_TYPE_CLOSE:
            return Close()

        len = struct.unpack('<i', data[4:8])
        if len(data) < 8 + size:
            return None
        odata = pkt[8:8+size]

        data = bp.decode(odata)
        if type == MESSAGE_TYPE_QUERY:
            return Query.from_receive(data)
        elif type == MESSAGE_TYPE_ANSWER:
            return Answer.from_receive(data)
        else:
            return None

class Welcome():

    def __repr__(self):
        return 'proxy.Welcome()'

class Close:

    def __repr__(self):
        return 'proxy.Close()'

class Query:
    def __init__(self, qid, service, method, params):
        self.qid = qid
        self.service = service
        self.method = method
        self.params = params

    def get_data(self):
        return Messager.data_for_query(self.qid, self.service, self.method, self.params)

    @staticmethod
    def from_receive(data):
        return Query(data[0], data[1], data[2], data[3])

    def __repr__(self):
        l = (self.qid, self.service, self.method, self.params)
        return 'proxy.Query(' + str(l) + ')'

class Answer:
    def __init__(self, qid, status, data):
        self.qid = qid
        self.status = status
        self.data = data

    def __repr__(self):
        return 'proxy.Answer(' + str((self.qid, self.status, self.data)) + ')'

    def get_data(self):
        return Messager.data_for_answer(self.qid, self.status, self.data)

    @staticmethod
    def from_receive(data):
        return Answer(data[0], data[1], data[2])

class ProxyError(Exception):
    def __init__(self, qid, status, params):
        self.status = status
        self.params = params
        self.qid = qid

    def __str__(self):
        return 'proxy.Error(' + str(self.params) + ')'

    def __repr__(self):
        return 'proxy.Error(' + str(self.params) + ')'

class Proxy(object):

    def __init__(self, end_point, timeout = 6000, welcome=True):
        self.end_point = end_point
        self.welcome = welcome

        self.service, self.host, self.port = parse_endpoint(end_point)

        self.timeout = timeout

        self._socket = None
        self.reset()

    def stop(self):

        self._running = False
        self._connected = False
        self._recv_thread = None
        self._reap_thread = None

        if self._socket:
            self._socket.close()
        self._socket = None

    def reset(self):
        self.last_id = 0
        self.pending = {}
        self._query_list = {}
        self._send_queue = Queue()
        self._recv_thread = None
        self._send_thread = gevent.spawn(self.send_fiber)
        self._reap_thread = gevent.spawn(self.timeout_fiber)
        self._connected = False
        self._running = True

    def next_qid(self):
        self.last_id +=1;
        return self.last_id;

    def connect(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self._socket.connect((self.host, self.port))

        ip, port = self._socket.getsockname()
        logger.get_logger().debug('address = %s:%d' % (ip, port))

        if self.welcome:
            type = Messager.receive_msg_type(self._socket)
            if type != MESSAGE_TYPE_WELCOME:
                raise Exception('Imcomplete message')

            logger.get_logger().debug('S.WELCOME')

    def unitive_send(self, data):
        self._socket.sendall(data)

    def unitive_recvfrom(self, recvlen):
        data = self._socket.recvfrom(recvlen)
        return data

    def send_quest(self, qid, method, params, twoway=True):
        if qid:
            self.pending[qid] = time.time()

        logger.get_logger().debug('CQ:%d service=%s method=%s', qid, self.service, method)

        self.unitive_send(Messager.data_for_query(qid, self.service, method, params))
        return qid

    def send_answer(self, qid, status, data):
        logger.get_logger().debug('CA:%d %d %s', qid, status, data)
        self.unitive_send(Messager.data_for_answer(qid, status, data))

    def receive_next_msg(self):
        message = Messager.receive_msg(self._socket)

        if isinstance(message, Query):
            logger.get_logger().debug('SQ:%d method=%s', message.qid, message.method)
        elif isinstance(message, Answer):
            logger.get_logger().debug('SA:%d status=%d', message.qid, message.status)

        if not message:
            raise Exception('Recieved Unknown packet')

        if isinstance(message, Answer):
            qid = message.qid
            if self.pending.get(qid):
                del self.pending[qid]
            status = message.status
            if status:
                raise ProxyError(qid, status, message.data)
        return message;

    def timeout_fiber(self):
        while True:
            try:
                now = time.time()
                for qid, ctime in self.pending.items():
                    timediff = (now - ctime) * 1000
                    if timediff > self.timeout:
                        exdict = {}
                        exdict['exception'] = 'QueryTimeoutError'
                        exdict['code'] = 1
                        exdict['message'] = 'query %d recv timeout ctime %d now %d proxy timeout %dms' % (qid, ctime, now, self.timeout)
                        exdict['detail'] = {}
                        if self._query_list.get(qid):
                            msg = Answer(qid, 1, exdict)
                            self._query_list[qid].set(msg)
                            del self._query_list[qid]
                            del self.pending[qid]
            except:
                logger.get_logger().error('proxy %s reap fiber exception %s', self.end_point, traceback.format_exc())

            gevent.sleep(3);

    def send_fiber(self):
        while self._running:
            qid = None
            try:
                qid, method, params = self._send_queue.get()
                if not self._connected:
                    self.connect()
                    self._connected = True
                    self._recv_thread = gevent.spawn(self.recv_fiber)
                self.send_quest(qid, method, params)
            except ProxyError as ex:
                self.close()
            except Exception as ex:
                self.close()
            finally:
                pass

    def recv_fiber(self):
        while self._running:
            try:
                msg = self.receive_next_msg()
                if msg == None:
                    raise params.EngineError('Received nothing')
                if isinstance(msg, Query):
                    raise params.EngineError('Recieved unexpected Query')
                elif isinstance(msg, Close):
                    self.close()
                elif isinstance(msg, Answer):
                    if self._query_list.get(msg.qid):
                        self._query_list[msg.qid].set(msg)
                        del self._query_list[msg.qid]
                    else:
                        logger.get_logger().error('proxy %s get unexpected answer %s. qid may be removed for timeout', self.end_point, msg)
                else:
                    raise params.EngineError('Unknown message')
            except ProxyError as ex:
                if self._query_list.get(ex.qid):
                    msg = Answer(ex.qid, ex.status, ex.params)
                    self._query_list[ex.qid].set(msg)
                    del self._query_list[ex.qid]
            except Exception as ex:
                self.close()
                break

    def close(self):
        logger.get_logger().info('close')
        if self._recv_thread:
            gevent.kill(self._recv_thread)
            self._recv_thread = None
        if self._socket:
            self._socket.close()
        self._socket = None
        self._connected = False

        self._fail_quests()

    def _build_except_info(self):
        exdict = {}
        exctype, exmsg, tb = sys.exc_info()
        exdict['exception'] = repr(exctype)
        exdict['code'] = 1
        exdict['message'] = repr(exmsg)
        exdict['raiser'] = self.end_point
        exdict['detail'] = {}
        exdict['detail']['what'] = repr(traceback.extract_tb(tb))
        return exdict

    def _fail_quests(self):
        exinfo = self._build_except_info()
        import traceback
        traceback.print_stack()
        logger.get_logger().error('proxy %s get error %s, fail current querys', self.end_point, exinfo)
        curr_req_items = self._query_list.items()
        self._query_list = {}
        for qid, result in curr_req_items:
            msg = Answer(qid, 1, exinfo)
            #AsyncResult set should cause context switch
            result.set(msg)

    def emit_query(self, qid, method, params):
        self._send_queue.put((qid, method, params))

    def request_many(self, reqs):
        """
        salvo simultaneous query many service and get result
        """
        local_quests = {}
        idx = 0
        while idx < len(reqs):
            req = reqs[idx]
            ar = AsyncResult()
            qid = self.next_qid()
            self.emit_query(qid, *req)
            self._query_list[qid] = ar
            local_quests[qid] = (idx, ar)
            idx += 1

        for qid, it in local_quests.items():
            idx, ar = it
            """
            async result will always get a response even if other query
            reply soon

            """
            msg = ar.get()
            if msg:
                if msg.status:
                    yield ProxyError(qid, msg.status, msg.params), idx
                else:
                    yield msg.data, idx

    def request(self, method, params, twoway = True):
        ar = AsyncResult()
        qid = 0;
        if twoway:
            qid = self.next_qid()
            self._query_list[qid] = ar

        self.emit_query(qid, method, params);

        if twoway:
            msg = ar.get()
            if msg.status:
                raise ProxyError(qid, msg.status, msg.data)
            return msg.data

def parse_endpoint(endpoint):
    x = endpoint.split('@')
    service = x[0].strip()

    z = x[1].split()
    host, port = z[0].split(':')

    if len(host.strip()) == 0:
        host = '127.0.0.1'

    port = int(port)
    return service, host, port
